/*
Colab colors

Forked from
https://github.com/highlightjs/highlight.js/blob/main/src/styles/default.css

TODO(epot): Is there a way to use Colab native `--syntax-class-color`,...
directly ?

*/

.hljs {
  /* --color: #f8f8f2; */
  --background-color: #f7f7f7;
  --property-color: #087db2;
  --string-color: #a31515;
  --number-color: #008000;
  --boolean-color: #0000ff;
  --null-color: #af00db;
  --preview-color: #888888;

  /* color: var(--color); */
  font-family: monospace;
  font-size: 14px;
}

html[theme=dark] .hljs {
  --background-color: #2c2c2c;
  --property-color: #6fb3d2;
  --string-color: #ce9178;
  --number-color: #b5cea8;
  --boolean-color: #569cd6;
  --null-color: #c586c0;
  --preview-color: #888888;
}


pre code.hljs {
  /* background-color: var(--background-color); */
  display: block;
  /* padding: 1em; */
}

.hljs-comment {
  color: var(--preview-color);
}

.hljs-keyword {
  color: var(--null-color);
}

.hljs-string {
  color: var(--string-color);
}

.hljs-number {
  color: var(--number-color);
}

.hljs-literal {
  color: var(--boolean-color);
}
